@@ -1,11 +1,11 @@
 /* Double-double arithmetic library
  *
  * Part of the functions are copied from the QD library for U.C. Berkeley
- * and licensed BSD.
+ * and licensed modified BSD (see QD-LICENSE.txt)
  *
  * Copyright (C) 2012 Yozo Hida, Xiaoye S. Li, David H. Bailey
  * Copyright (C) 2021 Markus Wallerberger and others
- * SPDX-License-Identifier: MIT, BSD
+ * SPDX-License-Identifier: MIT and Modified-BSD
  */
 #include "./dd_arith.h"
 #include <math.h>
@@ -23,7 +23,7 @@ ddouble _hypotqq_ordered(ddouble x, ddouble y)
 {
     // assume that x >= y >= 0
     // special cases
-    if (iszeroq(x))
+    if (iszeroq(y))
         return x;
 
     // if very large or very small, renormalize
@@ -63,9 +63,12 @@ ddouble sqrtq(ddouble a)
     return two_sum(ax, diff);
 }
 
-/* Inverse Factorials from 1/3!, 1/4!, asf. */
-static int _n_inv_fact = 15;
+/* Inverse Factorials from 1/0!, 1/1!, 1/2!, asf. */
+static int _n_inv_fact = 18;
 static const ddouble _inv_fact[] = {
+    {1.00000000000000000e+00, 0.00000000000000000e+00},
+    {1.00000000000000000e+00, 0.00000000000000000e+00},
+    {5.00000000000000000e-01, 0.00000000000000000e+00},
     {1.66666666666666657e-01, 9.25185853854297066e-18},
     {4.16666666666666644e-02, 2.31296463463574266e-18},
     {8.33333333333333322e-03, 1.15648231731787138e-19},
@@ -93,42 +96,41 @@ static const ddouble _inv_fact[] = {
  */
 static ddouble _exp_reduced(ddouble a, int *m)
 {
-    /* Strategy:  We first reduce the size of x by noting that
-     *
-     *     exp(k * r + m * log(2)) = 2^m * exp(r)^k
-     *
-     * where m and k are integers.  By choosing m appropriately
-     * we can make |k * r| <= log(2) / 2 = 0.347.
-     */
+    // Strategy:  We first reduce the size of x by noting that
+    //
+    //     exp(k * r + m * log(2)) = 2^m * exp(r)^k
+    //
+    // where m and k are integers.  By choosing m appropriately
+    // we can make |k * r| <= log(2) / 2 = 0.347.
     const double k = 512.0;
     const double inv_k = 1.0 / k;
     double mm = floor(a.hi / Q_LOG2.hi + 0.5);
     ddouble r = mul_pwr2(subqq(a, mulqd(Q_LOG2, mm)), inv_k);
     *m = (int)mm;
 
-    /* Now, evaluate exp(r) using the familiar Taylor series.  Reducing the
-     * argument substantially speeds up the convergence.  First, we compute
-     * terms of order 1 and 2 and add it to the sum
-     */
-    ddouble sum, term, rpower;
-    rpower = sqrq(r);
-    sum = addqq(r, mul_pwr2(rpower, 0.5));
+    // Now, evaluate exp(r) using the Taylor series, since reducing
+    // the argument substantially speeds up the convergence.  We omit order 0
+    // and start at order 1:
+    ddouble rpower = r;
+    ddouble term = r;
+    ddouble sum = term;
 
-    /* Next, compute terms of order 3 and up */
-    rpower = mulqq(rpower, r);
-    term = mulqq(rpower, _inv_fact[0]);
-    int i = 0;
-    do {
-        sum = addqq(sum, term);
-        rpower = mulqq(rpower, r);
-        ++i;
-        term = mulqq(rpower, _inv_fact[i]);
-    } while (fabs(term.hi) > inv_k * Q_EPS.hi && i < 5);
+    // Order 2
+    rpower = sqrq(r);
+    term = mul_pwr2(rpower, 0.5);
     sum = addqq(sum, term);
 
-    /* We now have that approximately exp(r) == 1 + sum.  Raise that to
-     * the m'th (512) power by squaring the binomial nine times
-     */
+    // Order 3 and up
+    for (int i = 3; i < 9; i++) {
+        rpower = mulqq(rpower, r);
+        term = mulqq(rpower, _inv_fact[i]);
+        sum = addqq(sum, term);
+        if (fabs(term.hi) <= inv_k * Q_EPS.hi)
+            break;
+    }
+
+    // We now have that approximately exp(r) == 1 + sum.  Raise that to
+    // the m'th (512) power by squaring the binomial nine times
     sum = addqq(mul_pwr2(sum, 2.0), sqrq(sum));
     sum = addqq(mul_pwr2(sum, 2.0), sqrq(sum));
     sum = addqq(mul_pwr2(sum, 2.0), sqrq(sum));
@@ -217,9 +219,6 @@ ddouble logq(ddouble a)
     return x;
 }
 
-static const ddouble _pi_16 =
-    {1.963495408493620697e-01, 7.654042494670957545e-18};
-
 /* Table of sin(k * pi/16) and cos(k * pi/16). */
 static const ddouble _sin_table[] = {
     {1.950903220161282758e-01, -7.991079068461731263e-18},
@@ -237,46 +236,46 @@ static const ddouble _cos_table[] = {
 
 static ddouble sin_taylor(ddouble a)
 {
+    // Use the Taylor series a - a^3/3! + a^5/5! + ...
     const double thresh = 0.5 * fabs(a.hi) * Q_EPS.hi;
-    ddouble r, s, t, x;
+    const ddouble minus_asquared = negq(sqrq(a));
 
-    if (iszeroq(a))
-        return Q_ZERO;
+    // First order:
+    ddouble apow = a;
+    ddouble term = a;
+    ddouble sum = a;
 
-    int i = 0;
-    x = negq(sqrq(a));
-    s = a;
-    r = a;
-    do {
-        r = mulqq(r, x);
-        t = mulqq(r, _inv_fact[i]);
-        s = addqq(s, t);
-        i += 2;
-    } while (i < _n_inv_fact && fabs(t.hi) > thresh);
-
-    return s;
+    // Subsequent orders:
+    for (int i = 3; i < _n_inv_fact; i += 2) {
+        apow = mulqq(apow, minus_asquared);
+        term = mulqq(apow, _inv_fact[i]);
+        sum = addqq(sum, term);
+        if (fabs(term.hi) <= thresh)
+            break;
+    }
+    return sum;
 }
 
 static ddouble cos_taylor(ddouble a)
 {
+    // Use Taylor series 1 - x^2/2! + x^4/4! + ...
     const double thresh = 0.5 * Q_EPS.hi;
-    ddouble r, s, t, x;
+    const ddouble minus_asquared = negq(sqrq(a));
 
-    if (iszeroq(a))
-        return Q_ONE;
+    // Zeroth and second order:
+    ddouble apow = minus_asquared;
+    ddouble term = mul_pwr2(apow, 0.5);
+    ddouble sum = adddq(1.0, term);
 
-    x = negq(sqrq(a));
-    r = x;
-    s = adddq(1.0, mul_pwr2(r, 0.5));
-    int i = 1;
-    do {
-        r = mulqq(r, x);
-        t = mulqq(r, _inv_fact[i]);
-        s = addqq(s, t);
-        i += 2;
-    } while (i < _n_inv_fact && fabs(t.hi) > thresh);
-
-    return s;
+    // From fourth order:
+    for (int i = 4; i < _n_inv_fact; i += 2) {
+        apow = mulqq(apow, minus_asquared);
+        term = mulqq(apow, _inv_fact[i]);
+        sum = addqq(sum, term);
+        if (fabs(term.hi) <= thresh)
+            break;
+    }
+    return sum;
 }
 
 static void sincos_taylor(ddouble a, ddouble *sin_a, ddouble *cos_a)
@@ -290,6 +289,35 @@ static void sincos_taylor(ddouble a, ddouble *sin_a, ddouble *cos_a)
     }
 }
 
+/**
+ * To compute 2pi-periodic function, we reduce the argument `a` by
+ * choosing integers z, -2 <= j <= 2 and -4 <= k <= 4 such that:
+ *
+ *   a == z * (2*pi) + j * (pi/2) + k * (pi/16) + t,
+ *
+ * where `abs(t) <= pi/32`.
+ */
+static ddouble mod_pi16(ddouble a, int *j, int *k)
+{
+    static const ddouble pi_16 =
+        {1.963495408493620697e-01, 7.654042494670957545e-18};
+
+    // approximately reduce modulo 2*pi
+    ddouble z = roundq(divqq(a, Q_2PI));
+    ddouble r = subqq(a, mulqq(Q_2PI, z));
+
+    // approximately reduce modulo pi/2
+    double q = floor(r.hi / Q_PI_2.hi + 0.5);
+    ddouble t = subqq(r, mulqd(Q_PI_2, q));
+    *j = (int)q;
+
+    // approximately reduce modulo pi/16.
+    q = floor(t.hi / pi_16.hi + 0.5);
+    t = subqq(t, mulqd(pi_16, q));
+    *k = (int)q;
+    return t;
+}
+
 ddouble sinq(ddouble a)
 {
     /* Strategy.  To compute sin(x), we choose integers a, b so that
@@ -306,18 +334,8 @@ ddouble sinq(ddouble a)
     if (iszeroq(a))
         return Q_ZERO;
 
-    // approximately reduce modulo 2*pi
-    ddouble z = roundq(divqq(a, Q_2PI));
-    ddouble r = subqq(a, mulqq(Q_2PI, z));
-
-    // approximately reduce modulo pi/2 and then modulo pi/16.
-    ddouble t;
-    double q = floor(r.hi / Q_PI_2.hi + 0.5);
-    t = subqq(r, mulqd(Q_PI_2, q));
-    int j = (int)q;
-    q = floor(t.hi / _pi_16.hi + 0.5);
-    t = subqq(t, mulqd(_pi_16, q));
-    int k = (int)q;
+    int j, k;
+    ddouble t = mod_pi16(a, &j, &k);
     int abs_k = abs(k);
 
     if (j < -2 || j > 2)
@@ -342,7 +360,7 @@ ddouble sinq(ddouble a)
 
     ddouble u = _cos_table[abs_k - 1];
     ddouble v = _sin_table[abs_k - 1];
-    ddouble sin_x, cos_x;
+    ddouble sin_x, cos_x, r;
     sincos_taylor(t, &sin_x, &cos_x);
     if (j == 0) {
         if (k > 0)
@@ -373,18 +391,8 @@ ddouble cosq(ddouble a)
     if (iszeroq(a))
         return Q_ONE;
 
-    // approximately reduce modulo 2*pi
-    ddouble z = roundq(divqq(a, Q_2PI));
-    ddouble r = subqq(a, mulqq(Q_2PI, z));
-
-    // approximately reduce modulo pi/2 and then modulo pi/16.
-    ddouble t;
-    double q = floor(r.hi / Q_PI_2.hi + 0.5);
-    t = subqq(r, mulqd(Q_PI_2, q));
-    int j = (int)q;
-    q = floor(t.hi / _pi_16.hi + 0.5);
-    t = subqq(t, mulqd(_pi_16, q));
-    int k = (int)q;
+    int j, k;
+    ddouble t = mod_pi16(a, &j, &k);
     int abs_k = abs(k);
 
     if (j < -2 || j > 2)
@@ -406,7 +414,7 @@ ddouble cosq(ddouble a)
         }
     }
 
-    ddouble sin_x, cos_x;
+    ddouble sin_x, cos_x, r;
     sincos_taylor(t, &sin_x, &cos_x);
     ddouble u = _cos_table[abs_k - 1];
     ddouble v = _sin_table[abs_k - 1];
@@ -449,22 +457,25 @@ ddouble sinhq(ddouble a)
         return mul_pwr2(subqq(ea, reciprocalq(ea)), 0.5);
     }
 
-    /* since a is small, using the above formula gives
-     * a lot of cancellation.  So use Taylor series.
-     */
-    ddouble s = a;
-    ddouble t = a;
-    ddouble r = sqrq(t);
-    double m = 1.0;
-    double thresh = fabs((a.hi) * Q_EPS.hi);
+    // When a is small, using the above formula gives a lot of cancellation.
+    // Use Taylor series: x + x^3/3! + x^5/5! + ...
+    const ddouble asquared = sqrq(a);
+    const double thresh = fabs(a.hi) * Q_EPS.hi;
 
-    do {
-        m += 2.0;
-        t = mulqq(t, r);
-        t = divqd(t, (m - 1) * m);
-        s = addqq(s, t);
-    } while (absq(t).hi > thresh);
-    return s;
+    // First order:
+    ddouble apower = a;
+    ddouble sum = a;
+    ddouble term = a;
+
+    // From third order:
+    for (int i = 3; i < _n_inv_fact; i += 2) {
+        apower = mulqq(apower, asquared);
+        term = mulqq(apower, _inv_fact[i]);
+        sum = addqq(sum, term);
+        if (fabs(term.hi) <= thresh)
+            break;
+    }
+    return sum;
 }
 
 ddouble coshq(ddouble a)
@@ -514,29 +525,11 @@ void sincosq(const ddouble a, ddouble *sin_a, ddouble *cos_a)
         return;
     }
 
-    // approximately reduce modulo 2*pi
-    ddouble z = nintq(divqq(a, Q_2PI));
-    ddouble r = subqq(a, mulqq(Q_2PI, z));
+    int j, k;
+    ddouble t = mod_pi16(a, &j, &k);
+    int abs_j = abs(j), abs_k = abs(k);
 
-    // approximately reduce module pi/2 and pi/16
-    ddouble t;
-    double q = floor(r.hi / Q_PI_2.hi + 0.5);
-    t = subqq(r, mulqd(Q_PI_2, q));
-    int j = (int)(q);
-    int abs_j = abs(j);
-    q = floor(t.hi / Q_PI_16.hi + 0.5);
-    t = subqq(t, mulqd(Q_PI_16, q));
-    int k = (int)(q);
-    int abs_k = abs(k);
-
-    if (abs_j > 2) {
-        // dd_real::error("(dd_real::sincos): Cannot reduce modulo pi/2.");
-        *cos_a = *sin_a = nanq();
-        return;
-    }
-
-    if (abs_k > 4) {
-        // dd_real::error("(dd_real::sincos): Cannot reduce modulo pi/16.");
+    if (abs_j > 2 || abs_k > 4) {
         *cos_a = *sin_a = nanq();
         return;
     }
@@ -580,21 +573,21 @@ void sincosq(const ddouble a, ddouble *sin_a, ddouble *cos_a)
 ddouble atan2qq(ddouble y, ddouble x)
 {
     /* Strategy: Instead of using Taylor series to compute
-        arctan, we instead use Newton's iteration to solve
-        the equation
-
-            sin(z) = y/r    or    cos(z) = x/r
-
-        where r = sqrt(x^2 + y^2).
-        The iteration is given by
-
-            z' = z + (y - sin(z)) / cos(z)          (for equation 1)
-            z' = z - (x - cos(z)) / sin(z)          (for equation 2)
-
-        Here, x and y are normalized so that x^2 + y^2 = 1.
-        If |x| > |y|, then first iteration is used since the
-        denominator is larger.  Otherwise, the second is used.
-    */
+     * arctan, we instead use Newton's iteration to solve
+     * the equation
+     *
+     *     sin(z) = y/r    or    cos(z) = x/r
+     *
+     * where r = sqrt(x^2 + y^2).
+     * The iteration is given by
+     *
+     *     z' = z + (y - sin(z)) / cos(z)          (for equation 1)
+     *     z' = z - (x - cos(z)) / sin(z)          (for equation 2)
+     *
+     * Here, x and y are normalized so that x^2 + y^2 = 1.
+     * If |x| > |y|, then first iteration is used since the
+     * denominator is larger.  Otherwise, the second is used.
+     */
     if (iszeroq(x) && iszeroq(y))
         return Q_ZERO;
     if (iszeroq(x))
