@@ -95,7 +95,65 @@ def rrqr(A, tol=5e-32, reflectors=False):
     return Q, R, jpvt
 
 
-def bidiag(A, reflectors=False):
+def svd(A, full_matrices=True):
+    """Truncated singular value decomposition.
+
+    Decomposes a `(m, n)` matrix `A` into the product:
+
+        A == U @ (s[:,None] * VT)
+
+    where `U` is a `(m, k)` matrix with orthogonal columns, `VT` is a `(k, n)`
+    matrix with orthogonal rows and `s` are the singular values, a set of `k`
+    nonnegative numbers in non-ascending order and `k = min(m, n)`.
+    """
+    A = np.asarray(A)
+    m, n = A.shape
+    if m < n:
+        U, s, VT = svd(A.T, full_matrices)
+        return VT.T, s, U.T
+
+    Q, B, RT = bidiag(A)
+    for _ in range(20 * n):
+        if svd_bidiag_step(Q, B, RT):
+            break
+    else:
+        warn("Did not converge")
+
+    U, s, VH = svd_normalize(Q, B.diagonal(), RT)
+    if not full_matrices:
+        U = U[:,:n]
+    return U, s, VH
+
+
+def svd_trunc(A, tol=5e-32, method='jacobi', max_iter=20):
+    """Truncated singular value decomposition.
+
+    Decomposes a `(m, n)` matrix `A` into the product:
+
+        A == U @ (s[:,None] * VT)
+
+    where `U` is a `(m, k)` matrix with orthogonal columns, `VT` is a `(k, n)`
+    matrix with orthogonal rows and `s` are the singular values, a set of `k`
+    nonnegative numbers in non-ascending order.  The SVD is truncated in the
+    sense that singular values below `tol` are discarded.
+    """
+    # RRQR is an excellent preconditioner for Jacobi.  One should then perform
+    # Jacobi on RT
+    Q, R, p = rrqr(A, tol)
+    if method == 'jacobi':
+        U, s, VT = svd_jacobi(R.T, tol, max_iter)
+    elif method == 'golub-kahan':
+        U, s, VT = svd(R.T, full_matrices=False)
+    else:
+        raise ValueError("invalid method")
+
+    # Reconstruct A from QRs
+    U_A = Q @ VT.T
+    VT_B = U.T[:, p.argsort()]
+    return U_A, s, VT_B
+
+
+def bidiag(A, reflectors=False, force_structure=False):
     """Biadiagonalizes an arbitray rectangular matrix.
 
     Decomposes a `(m, n)` matrix `A` into the product:
@@ -123,29 +181,26 @@ def bidiag(A, reflectors=False):
     for j in range(n-2, rq):
         householder_update(B[j:,j:], Q[j:,j:])
 
+    if force_structure:
+        d = B.diagonal().copy()
+        e = B.diagonal(1).copy()
+        B[...] = 0
+        i = np.arange(n)
+        B[i, i] = d
+        B[i[:-1], i[:-1]+1] = e
     if not reflectors:
         Q = householder_apply(Q, np.eye(m, dtype=B.dtype))
         R = householder_apply(R, np.eye(n, dtype=B.dtype))
     return Q, B, R.T
 
 
-def svd_trunc(A, tol=5e-32, max_iter=20):
-    """Truncated singular value decomposition.
+def svd_jacobi(A, tol=5e-32, max_iter=20):
+    """Singular value decomposition using Jacobi rotations."""
+    U = A.copy()
+    m, n = U.shape
+    if m < n:
+        raise RuntimeError("expecting tall matrix")
 
-    Decomposes a `(m, n)` matrix `A` into the product:
-
-        A == U @ (s[:,None] * VT)
-
-    where `U` is a `(m, k)` matrix with orthogonal columns, `VT` is a `(k, n)`
-    matrix with orthogonal rows and `s` are the singular values, a set of `k`
-    nonnegative numbers in non-ascending order.  The SVD is truncated in the
-    sense that singular values below `tol` are discarded.
-    """
-    # RRQR is an excellent preconditioner for Jacobi.  One should then perform
-    # Jacobi on RT
-    Q, R, p = rrqr(A, tol)
-    U = R.T.copy()
-    _, n = U.shape
     VT = np.eye(n, dtype=U.dtype)
     offd = np.empty((), ddouble)
 
@@ -159,11 +214,7 @@ def svd_trunc(A, tol=5e-32, max_iter=20):
 
     s = norm(U.T)
     U = U / s
-
-    # Reconstruct A from QRs
-    U_A = Q @ VT.T
-    VT_B = U.T[:, p.argsort()]
-    return U_A, s, VT_B
+    return U, s, VT
 
 
 def householder_update(A, Q):
@@ -195,3 +246,73 @@ def householder_apply(H, Q):
         w = -beta * (Qpart.T @ v)
         rank1update(Qpart, v, w, out=Qpart)
     return Q
+
+
+def svd_normalize(U, d, VH):
+    """Given a SVD-like decomposition, normalize"""
+    # Invert
+    n = d.size
+    VH[np.signbit(d)] = -VH[np.signbit(d)]
+    d = np.abs(d)
+
+    # Sort
+    order = np.argsort(d)[::-1]
+    d = d[order]
+    VH = VH[order]
+    U = U.copy()
+    U[:,:n] = U[:,order]
+    return U, d, VH
+
+
+def svd_bidiag_step(Q, B, RT):
+    """Single SVD step for a bidiagonal matrix"""
+    d = B.diagonal().copy()
+    e = np.hstack([B.diagonal(1), 0.0])
+
+    p, q = bidiag_partition(d, e)
+    if q <= 1:
+        return True
+
+    d_part = d[p:q]
+    e_part = e[p:q]
+    rot = np.empty((d_part.size, 4), d.dtype)
+    _dd_linalg.golub_kahan_chase(d_part, e_part, out=(d_part, e_part, rot))
+
+    i = np.arange(p, q)
+    B[i, i] = d_part
+    B[i[:-1], i[:-1]+1] = e_part[:-1]
+
+    rot_Q = rot[:, 2:]
+    rot_R = rot[:, :2]
+    QT_part = Q[:, p:q].T
+    RT_part = RT[p:q, :]
+    _dd_linalg.givens_seq(rot_Q, QT_part, out=QT_part)
+    _dd_linalg.givens_seq(rot_R, RT_part, out=RT_part)
+    return False
+
+
+def bidiag_partition(d, e, eps=5e-32):
+    """Partition bidiagonal matrix into blocks for implicit QR.
+
+    Return `p,q` which partions a bidiagonal `B` matrix into three blocks:
+
+      - B[0:p, 0:p], an arbitrary bidiaonal matrix
+      - B[p:q, p:q], a matrix with all off-diagonal elements nonzero
+      - B[q:,  q:],  a diagonal matrix
+    """
+    abs_e = np.abs(e)
+    abs_d = np.abs(d)
+    e_zero = abs_e <= eps * (abs_d + abs_e)
+    e[e_zero] = 0
+
+    q = _find_last(~e_zero) + 1
+    if q <= 0:
+        return 0, 0
+    p = _find_last(e_zero[:q]) + 1
+    return p, q + 1
+
+
+def _find_last(a, axis=-1):
+    a = a.astype(bool)
+    maxloc = a.shape[axis] - 1 - a[::-1].argmax(axis)
+    return np.where(a[maxloc], maxloc, -1)
